# Contents in this file are referenced from the sphinx-generated docs.
# "magictoken" is used for markers as beginning and ending of example text.

import unittest
from numba.tests.support import captured_stdout, skip_parfors_unsupported
from numba import set_parallel_chunksize
from numba.tests.support import TestCase


@skip_parfors_unsupported
class ChunksizeExamplesTest(TestCase):

    _numba_parallel_test_ = False

    def setUp(self):
        set_parallel_chunksize(0)

    def tearDown(self):
        set_parallel_chunksize(0)

    def test_unbalanced_example(self):
        with captured_stdout():
            # magictoken.ex_unbalanced.begin
            from numba import (njit,
                               prange,
                               )
            import numpy as np

            @njit(parallel=True)
            def func1():
                n = 100
                vals = np.empty(n)
                # The work in each iteration of the following prange
                # loop is proportional to its index.
                for i in prange(n):
                    cur = i + 1
                    for j in range(i):
                        if cur % 2 == 0:
                            cur //= 2
                        else:
                            cur = cur * 3 + 1
                    vals[i] = cur
                return vals

            result = func1()
            # magictoken.ex_unbalanced.end
            self.assertPreciseEqual(result, func1.py_func())

    def test_chunksize_manual(self):
        with captured_stdout():
            # magictoken.ex_chunksize_manual.begin
            from numba import (njit,
                               prange,
                               set_parallel_chunksize,
                               get_parallel_chunksize,
                               )

            @njit(parallel=True)
            def func1(n):
                acc = 0
                print(get_parallel_chunksize()) # Will print 4.
                for i in prange(n):
                    print(get_parallel_chunksize()) # Will print 0.
                    acc += i
                print(get_parallel_chunksize()) # Will print 4.
                return acc

            @njit(parallel=True)
            def func2(n):
                acc = 0
                # This version gets the previous chunksize explicitly.
                old_chunksize = get_parallel_chunksize()
                set_parallel_chunksize(8)
                for i in prange(n):
                    acc += i
                set_parallel_chunksize(old_chunksize)
                return acc

            # This version saves the previous chunksize as returned
            # by set_parallel_chunksize.
            old_chunksize = set_parallel_chunksize(4)
            result1 = func1(12)
            result2 = func2(12)
            result3 = func1(12)
            set_parallel_chunksize(old_chunksize)
            # magictoken.ex_chunksize_manual.end
            self.assertPreciseEqual(result1, func1.py_func(12))
            self.assertPreciseEqual(result2, func2.py_func(12))
            self.assertPreciseEqual(result3, func1.py_func(12))

    def test_chunksize_with(self):
        with captured_stdout():
            # magictoken.ex_chunksize_with.begin
            from numba import njit, prange, parallel_chunksize

            @njit(parallel=True)
            def func1(n):
                acc = 0
                for i in prange(n):
                    acc += i
                return acc

            @njit(parallel=True)
            def func2(n):
                acc = 0
                with parallel_chunksize(8):
                    for i in prange(n):
                        acc += i
                return acc

            with parallel_chunksize(4):
                result1 = func1(12)
                result2 = func2(12)
                result3 = func1(12)
            # magictoken.ex_chunksize_with.end
            self.assertPreciseEqual(result1, func1.py_func(12))
            self.assertPreciseEqual(result2, func2.py_func(12))
            self.assertPreciseEqual(result3, func1.py_func(12))


if __name__ == '__main__':
    unittest.main()



"""
Implementation of method overloads for Generator objects.
"""

import numpy as np
from numba.core import types
from numba.core.extending import overload_method
from numba.np.numpy_support import as_dtype, from_dtype
from numba.np.random.generator_core import next_float, next_double
from numba.np.numpy_support import is_nonelike
from numba.core.errors import TypingError
from numba.core.types.containers import Tuple, UniTuple


def _get_proper_func(func_32, func_64, dtype, dist_name="the given"):
    """
        Most of the standard NumPy distributions that accept dtype argument
        only support either np.float32 or np.float64 as dtypes.

        This is a helper function that helps Numba select the proper underlying
        implementation according to provided dtype.
    """
    if isinstance(dtype, types.Omitted):
        dtype = dtype.value

    if not isinstance(dtype, types.Type):
        dt = np.dtype(dtype)
        nb_dt = from_dtype(dt)
        np_dt = dtype
    else:
        nb_dt = dtype
        np_dt = as_dtype(nb_dt)

    np_dt = np.dtype(np_dt)

    if np_dt == np.float32:
        next_func = func_32
    elif np_dt == np.float64:
        next_func = func_64
    else:
        raise TypingError(
            f"Unsupported dtype {np_dt} for {dist_name} distribution"
        )

    return next_func, nb_dt


# Overload the Generator().random()
@overload_method(types.NumPyRandomGeneratorType, 'random')
def NumPyRandomGeneratorType_random(inst, size=None, dtype=np.float64):
    dist_func, nb_dt = _get_proper_func(next_float, next_double,
                                        dtype, "random")
    if isinstance(size, types.Omitted):
        size = size.value

    if is_nonelike(size):
        def impl(inst, size=None, dtype=np.float64):
            return nb_dt(dist_func(inst.bit_generator))
        return impl
    else:
        assert any([isinstance(size, UniTuple) and
                   isinstance(size.dtype, types.Integer),
                   isinstance(size, Tuple) and size.count == 0,
                   isinstance(size, types.Integer)]
                   ), "Size argument either of None,"\
            + " an integer, an empty tuple or a tuple of integers"

        def impl(inst, size=None, dtype=np.float64):
            out = np.empty(size, dtype=dtype)
            for i in np.ndindex(size):
                out[i] = dist_func(inst.bit_generator)
            return out
        return impl



# "magictoken" is used for markers as beginning and ending of example text.

import unittest
from numba.tests.support import captured_stdout, override_config


class DocsLLVMPassTimings(unittest.TestCase):

    def test_pass_timings(self):
        with override_config('LLVM_PASS_TIMINGS', True):
            with captured_stdout() as stdout:
                # magictoken.ex_llvm_pass_timings.begin
                import numba

                @numba.njit
                def foo(n):
                    c = 0
                    for i in range(n):
                        for j in range(i):
                            c += j
                    return c

                foo(10)
                md = foo.get_metadata(foo.signatures[0])
                print(md['llvm_pass_timings'])
                # magictoken.ex_llvm_pass_timings.end
            self.assertIn("Finalize object", stdout.getvalue())


if __name__ == '__main__':
    unittest.main()
